import logging
import os
import warnings
from collections import UserDict
from contextlib import contextmanager
from typing import Any, Dict, List, MutableSequence, Optional, Sequence, Tuple, TypeVar, Union

import torch
import tqdm
from packaging import version
from pie_core import AnnotationPipeline, Document, TaskEncoding, TaskModule
from torch import Tensor
from torch.utils.data import DataLoader
from transformers.utils import ModelOutput

from pytorch_ie.dataset import TaskEncodingDataset
from pytorch_ie.model import AutoPyTorchIEModel, PyTorchIEModel


class InplaceNotSupportedException(Exception):
    pass


TaskOutput = TypeVar("TaskOutput")

logger = logging.getLogger(__name__)


# TODO: use torch.get_autocast_dtype when available
def get_autocast_dtype(device_type: str):
    if device_type == "cuda":
        return torch.float16
    elif device_type == "cpu":
        return torch.bfloat16
    else:
        raise ValueError(f"Unsupported device type for half precision autocast: {device_type}")


@AnnotationPipeline.register()
class PyTorchIEPipeline(AnnotationPipeline[PyTorchIEModel, TaskModule]):
    """
    The Pipeline class is the class from which all pipelines inherit. Refer to this class for methods shared across
    different pipelines.

    Base class implementing pipelined operations. Pipeline workflow is defined as a sequence of the following
    operations:

        Input -> Tokenization -> Model Inference -> Post-Processing (task dependent) -> Output

    Pipeline supports running on CPU or GPU through the device argument (see below).

    Args:
        model (:class:`~pytorch_ie.PyTorchIEModel`):
            The deep learning model to use for the pipeline.
        taskmodule (:class:`~pytorch_ie.TaskModule`): The taskmodule to use for encoding
            and decoding the documents.
        device (:obj:`Union[int, str]`, `optional`, defaults to :obj:`"cpu"`):
            The device to run the pipeline on. This can be a CPU device (:obj:`"cpu"`), a GPU
            device (:obj:`"cuda"`) or a specific GPU device (:obj:`"cuda:X"`, where :obj:`X`
            is the index of the GPU).
        half_precision_model (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not to use half precision model. This can be set to :obj:`True` to reduce
            the memory usage of the model. If set to :obj:`True`, the model will be cast to
            :obj:`torch.float16` on supported devices.
    """

    # TODO: This is required for backward compatibility because all models so far are annotated with
    # @PyTorchIEModel.register(). However, this prevents AutoAnnotationPipeline.from_pretrained() and .from_config()
    # from working correctly because it still has auto_model_class = AutoModel. We could define a class
    # AutoPyTorchIEPipeline(AutoAnnotationPipeline) with auto_model_class = AutoPyTorchIEModel, but
    # this mitigates the purpose of the AutoModel class. In the future, we should remove this
    # and register all models with @Model.register() instead (but this will break backwards compatibility).
    auto_model_class = AutoPyTorchIEModel

    default_input_names = None

    def __init__(
        self,
        device: Union[int, str] = "cpu",
        half_precision_model: bool = False,
        **kwargs,
    ):
        (
            self._preprocess_params,
            self._dataloader_params,
            self._forward_params,
            self._postprocess_params,
            remaining_kwargs,
        ) = self._sanitize_parameters(**kwargs)

        # we need to pass the remaining arguments (i.e., model, taskmodule, is_from_pretrained) to the parent class
        super().__init__(**remaining_kwargs)

        device_str = (
            ("cpu" if device < 0 else f"cuda:{device}") if isinstance(device, int) else device
        )
        self.device = torch.device(device_str)

        # Module.to() returns just self, but moved to the device. This is not correctly
        # reflected in typing of PyTorch.
        self.model: PyTorchIEModel = self.model.to(self.device)  # type: ignore
        if half_precision_model:
            self.model = self.model.to(dtype=get_autocast_dtype(self.device.type))

        self.call_count = 0

    def transform(self, X):
        """
        Scikit / Keras interface to transformers' pipelines. This method will forward to __call__().
        """
        return self(X=X)

    def predict(self, X):
        """
        Scikit / Keras interface to transformers' pipelines. This method will forward to __call__().
        """
        return self(X=X)

    @contextmanager
    def device_placement(self):
        """
        Context Manager allowing tensor allocation on the user-specified device in framework agnostic way.

        Returns:
            Context manager

        Examples::

            # Explicitly ask for tensor allocation on CUDA device :0
            pipe = pipeline(..., device=0)
            with pipe.device_placement():
                # Every framework specific tensor allocation will be done on the request device
                output = pipe(...)
        """
        if self.device.type == "cuda":
            torch.cuda.set_device(self.device)

        yield

    def ensure_tensor_on_device(self, **inputs):
        """
        Ensure PyTorch tensors are on the specified device.

        Args:
            inputs (keyword arguments that should be :obj:`torch.Tensor`, the rest is ignored): The tensors to place on :obj:`self.device`.
            Recursive on lists **only**.

        Return:
            :obj:`Dict[str, torch.Tensor]`: The same as :obj:`inputs` but on the proper device.
        """
        return self._ensure_tensor_on_device(inputs, self.device)

    def _ensure_tensor_on_device(self, inputs, device):
        if isinstance(inputs, ModelOutput):
            return type(inputs)(
                {
                    name: self._ensure_tensor_on_device(tensor, device)
                    for name, tensor in inputs.items()
                }
            )
        elif isinstance(inputs, dict):
            return {
                name: self._ensure_tensor_on_device(tensor, device)
                for name, tensor in inputs.items()
            }
        elif isinstance(inputs, UserDict):
            return UserDict(
                {
                    name: self._ensure_tensor_on_device(tensor, device)
                    for name, tensor in inputs.items()
                }
            )
        elif isinstance(inputs, list):
            return [self._ensure_tensor_on_device(item, device) for item in inputs]
        elif isinstance(inputs, tuple):
            return tuple(self._ensure_tensor_on_device(item, device) for item in inputs)
        elif isinstance(inputs, torch.Tensor):
            return inputs.to(device)
        else:
            return inputs

    def _sanitize_parameters(
        self, **pipeline_parameters
    ) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        """
        _sanitize_parameters will be called with any excessive named arguments from either `__init__` or `__call__`
        methods. It should return 4 dictionaries of the resolved parameters used by the various `preprocess`,
        `get_dataloader`, `forward` and `postprocess` methods. Do not fill dictionaries if the caller didn't specify
        a kwargs. This lets you keep defaults in function signatures, which is more "natural".

        It is not meant to be called directly, it will be automatically called and the final parameters resolved by
        `__init__` and `__call__`
        """
        preprocess_parameters = {}
        dataloader_params = {}
        forward_parameters = {}
        postprocess_parameters: Dict[str, Any] = {}

        # set preprocess parameters
        for p_name in ["document_batch_size"]:
            if p_name in pipeline_parameters:
                preprocess_parameters[p_name] = pipeline_parameters.pop(p_name)

        # set forward parameters
        for p_name in ["show_progress_bar", "fast_dev_run", "half_precision_ops"]:
            if p_name in pipeline_parameters:
                forward_parameters[p_name] = pipeline_parameters.pop(p_name)

        # set dataloader parameters
        for p_name in ["batch_size", "num_workers"]:
            if p_name in pipeline_parameters:
                dataloader_params[p_name] = pipeline_parameters.pop(p_name)

        # set postprocess parameters
        for p_name in ["inplace"]:
            if p_name in pipeline_parameters:
                postprocess_parameters[p_name] = pipeline_parameters.pop(p_name)

        return (
            preprocess_parameters,
            dataloader_params,
            forward_parameters,
            postprocess_parameters,
            pipeline_parameters,
        )

    def preprocess(
        self,
        documents: Sequence[Document],
        document_batch_size: Optional[int] = None,
        **preprocess_parameters: Dict,
    ) -> Sequence[TaskEncoding]:
        """
        Preprocess will take the `input_` of a specific pipeline and return a dictionary of everything necessary for
        `_forward` to run properly. It should contain at least one tensor, but might have arbitrary other items.
        """

        encodings = self.taskmodule.encode(
            documents,
            encode_target=False,
            as_task_encoding_sequence=True,
            document_batch_size=document_batch_size,
        )
        if not isinstance(encodings, Sequence):
            raise TypeError("preprocess has to return a sequence")
        return encodings

    def _forward(
        self, input_tensors: Tuple[Dict[str, Tensor], Any, Any, Any], **forward_parameters: Dict
    ) -> Dict:
        """
        _forward will receive the prepared dictionary from `preprocess` and run it on the model. This method might
        involve the GPU or the CPU and should be agnostic to it. Isolating this function is the reason for `preprocess`
        and `postprocess` to exist, so that the hot path, this method generally can run as fast as possible.

        It is not meant to be called directly, `forward` is preferred. It is basically the same but contains additional
        code surrounding `_forward` making sure tensors and models are on the same device, disabling the training part
        of the code (leading to faster inference).
        """
        inputs = input_tensors[0]
        return self.model.predict(inputs, **forward_parameters)

    def postprocess(
        self,
        model_inputs: Sequence[TaskEncoding],
        model_outputs: Sequence[TaskOutput],
        **postprocess_parameters,
    ) -> Sequence[Document]:
        """
        Postprocess will receive the model inputs and (unbatched) model outputs and reformat them into
        something more friendly. Generally it will output a list of documents.
        """
        # This creates annotations from the model outputs and attaches them to the correct documents.
        return self.taskmodule.decode(
            task_encodings=model_inputs,
            task_outputs=model_outputs,
            **postprocess_parameters,
        )

    def get_inference_context(self):
        inference_context = (
            torch.inference_mode
            if version.parse(torch.__version__) >= version.parse("1.9.0")
            else torch.no_grad
        )
        return inference_context

    def forward(self, model_inputs, **forward_params):
        with self.device_placement():
            inference_context = self.get_inference_context()
            with inference_context():
                model_inputs = self._ensure_tensor_on_device(model_inputs, device=self.device)
                model_outputs = self._forward(model_inputs, **forward_params)
                model_outputs = self._ensure_tensor_on_device(
                    model_outputs, device=torch.device("cpu")
                )
        return model_outputs

    def get_dataloader(
        self,
        model_inputs: Sequence[TaskEncoding],
        batch_size: int = 1,
        num_workers: int = 8,
        **kwargs,
    ):
        dataloader: DataLoader[TaskEncoding] = DataLoader(
            TaskEncodingDataset(model_inputs),
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            collate_fn=self.taskmodule.collate,
            **kwargs,
        )

        return dataloader

    def __call__(
        self,
        documents: Union[Document, Sequence[Document]],
        *args,
        **kwargs,
    ) -> Union[Document, Sequence[Document]]:
        """
        The __call__ method is the entry point for the pipeline. It will run the pipeline workflow in the following
        order:

                1. Encode the documents
                2. Run the model forward pass(es) on the encodings
                3. Combine the model outputs with the input encodings and integrate them back into the documents

        Args:
            documents (:obj:`Union[Document, Sequence[Document]]`): The documents to process. If a single document is
                passed, the output will be a single document. If a list of documents is passed, the output will be a
                list of documents.
            document_batch_size (:obj:`int`, `optional`): The batch size to use for encoding the documents with the
                taskmodule. If not provided, the default batch size of the taskmodule will be used.
            show_progress_bar (:obj:`bool`, `optional`, defaults to :obj:`False`): Whether or not to show a progress bar
                during inference.
            fast_dev_run (:obj:`bool`, `optional`, defaults to :obj:`False`): Whether or not to run a fast development
                run. If set to :obj:`True`, only the first two model inputs will be processed.
            half_precision_ops (:obj:`bool`, `optional`, defaults to :obj:`False`): Whether or not to use half
                precision operations. If set to :obj:`True`, the model will be run with half precision operations
                via :obj:`torch.autocast`.
            batch_size (:obj:`int`, `optional`, defaults to :obj:`1`): The batch size to use for the dataloader. If not
                provided, a batch size of 1 will be used.
            num_workers (:obj:`int`, `optional`, defaults to :obj:`8`): The number of workers to use for the dataloader.
                If not provided, 8 workers will be used.
            inplace (:obj:`bool`, `optional`, defaults to :obj:`True`): Whether or not to modify the input documents
                in place. Requires the input to be a mutable sequence of documents or a single document.

        Note that all the arguments except `documents` can be set in the `__init__` method and/or overridden in the
        `__call__` method.

        Returns:
            :obj:`Union[Document, Sequence[Document]]`: The processed documents. If a single document was passed, a
            single document will be returned. If a list of documents was passed, a list of documents will be returned.
        """
        if args:
            logger.warning(f"Ignoring args: {args}")
        (
            preprocess_params,
            dataloader_params,
            forward_params,
            postprocess_params,
            remaining_kwargs,
        ) = self._sanitize_parameters(**kwargs)
        if remaining_kwargs:
            logger.warning(f"Ignoring remaining kwargs: {remaining_kwargs}")

        in_place: bool = postprocess_params.get("inplace", True)
        if in_place and not isinstance(documents, (MutableSequence, Document)):
            raise InplaceNotSupportedException(
                "Immutable sequences of Documents (such as Datasets) can't be modified in place. Please set inplace=False."
            )

        if "TOKENIZERS_PARALLELISM" not in os.environ:
            logger.info(
                "Disabling tokenizer parallelism, we're using DataLoader multithreading already"
            )
            os.environ["TOKENIZERS_PARALLELISM"] = "false"

        # Fuse __init__ params and __call__ params without modifying the __init__ ones.
        preprocess_params = {**self._preprocess_params, **preprocess_params}
        dataloader_params = {**self._dataloader_params, **dataloader_params}
        forward_params = {**self._forward_params, **forward_params}
        postprocess_params = {**self._postprocess_params, **postprocess_params}

        single_document = False
        if isinstance(documents, Document):
            single_document = True
            documents = [documents]

        # This creates encodings from the documents. It modifies the documents and may produce multiple entries per
        # document.
        model_inputs = self.preprocess(documents, **preprocess_params)
        if forward_params.pop("fast_dev_run", False):
            warnings.warn(
                "Execute a fast dev run, only the first two model inputs will be processed."
            )
            model_inputs = model_inputs[:2]
        # Create a dataloader from the model inputs. This uses taskmodule.collate().
        dataloader = self.get_dataloader(model_inputs=model_inputs, **dataloader_params)

        show_progress_bar = forward_params.pop("show_progress_bar", False)
        half_precision_ops = forward_params.pop("half_precision_ops", False)

        # Torch documentation recommends: "When entering an autocast-enabled region, Tensors may be any type.
        # You should not call half() or bfloat16() on your model(s) or inputs when using autocasting."
        # (see https://docs.pytorch.org/docs/stable/amp.html#torch.autocast). So show a warning in this case.
        if half_precision_ops:
            if self.model.dtype == get_autocast_dtype(self.device.type):
                logger.warning(
                    "Using half precision operations with a model already in half precision. "
                    "This is not recommended, as it may lead to unexpected results."
                )

        model_outputs: List = []
        with torch.no_grad():
            with torch.autocast(device_type=self.device.type, enabled=half_precision_ops):
                for batch in tqdm.tqdm(
                    dataloader, desc="inference", disable=not show_progress_bar
                ):
                    output = self.forward(batch, **forward_params)
                    processed_output = self.taskmodule.unbatch_output(output)
                    model_outputs.extend(processed_output)

        assert len(model_inputs) == len(
            model_outputs
        ), f"length mismatch: len(model_inputs) [{len(model_inputs)}] != len(model_outputs) [{len(model_outputs)}]"

        documents = self.postprocess(
            model_inputs=model_inputs,
            model_outputs=model_outputs,
            **postprocess_params,
        )
        if single_document:
            return documents[0]
        else:
            return documents


# kept for backward compatibility
Pipeline = PyTorchIEPipeline



from typing import Any, Dict, Generic, Iterable, Iterator, Optional, Sequence, TypeVar, Union

from pie_core import Document, TaskEncoding, TaskModule
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from pytorch_ie.dataset import IterableTaskEncodingDataset, TaskEncodingDataset

DocumentType = TypeVar("DocumentType", bound=Document)
InputEncoding = TypeVar("InputEncoding")
TargetEncoding = TypeVar("TargetEncoding")


class PieDataModule(LightningDataModule, Generic[DocumentType, InputEncoding, TargetEncoding]):
    """A simple LightningDataModule for PIE document datasets.

    A DataModule implements 5 key methods:
        - prepare_data (things to do on 1 GPU/TPU, not on every GPU/TPU in distributed mode)
        - setup (things to do on every accelerator in distributed mode)
        - train_dataloader (the training dataloader)
        - val_dataloader (the validation dataloader(s))
        - test_dataloader (the test dataloader(s))

    This allows you to share a full dataset without explaining how to download,
    split, transform and process the data.

    Read the docs:
        https://pytorch-lightning.readthedocs.io/en/latest/extensions/datamodules.html
    """

    def __init__(
        self,
        taskmodule: TaskModule[DocumentType, InputEncoding, TargetEncoding, Any, Any, Any],
        dataset: Dict[str, Sequence[DocumentType]],
        data_config_path: Optional[str] = None,
        train_split: Optional[str] = "train",
        val_split: Optional[str] = "validation",
        test_split: Optional[str] = "test",
        show_progress_for_encode: bool = False,
        **dataloader_kwargs,
    ):
        super().__init__()

        self.taskmodule = taskmodule
        self.config_path = data_config_path
        self.dataset = dataset
        self.train_split = train_split
        self.val_split = val_split
        self.test_split = test_split
        self.show_progress_for_encode = show_progress_for_encode
        self.dataloader_kwargs = dataloader_kwargs

        self._data: Dict[
            str,
            Union[
                TaskEncodingDataset[TaskEncoding[DocumentType, InputEncoding, TargetEncoding]],
                IterableTaskEncodingDataset[
                    TaskEncoding[DocumentType, InputEncoding, TargetEncoding]
                ],
            ],
        ] = {}

    def get_split_size(self, split: str) -> int:
        data = self._data.get(split, None)
        if data is None:
            raise ValueError(f"can not get {split} size if setup() was not yet called")
        if isinstance(data, IterableTaskEncodingDataset):
            raise TypeError("IterableTaskEncodingDataset has no length")
        return len(data)

    @property
    def num_train(self) -> int:
        if self.train_split is None:
            raise ValueError("no train split assigned")
        return self.get_split_size(self.train_split)

    @property
    def num_val(self) -> int:
        if self.val_split is None:
            raise ValueError("no val split assigned")
        return self.get_split_size(self.val_split)

    @property
    def num_test(self) -> int:
        if self.test_split is None:
            raise ValueError("no test split assigned")
        return self.get_split_size(self.test_split)

    def encode_documents(
        self, documents: Iterable[DocumentType]
    ) -> Union[TaskEncodingDataset, IterableTaskEncodingDataset]:
        task_encodings = self.taskmodule.encode(
            documents,
            encode_target=True,
            show_progress=self.show_progress_for_encode,
        )
        if isinstance(task_encodings, Sequence):
            return TaskEncodingDataset(task_encodings)
        elif isinstance(task_encodings, Iterator):
            return IterableTaskEncodingDataset(task_encodings)
        else:
            raise TypeError(
                f"task_encodings should be a Sequence or Iterator, but got {type(task_encodings)}"
            )

    def setup(self, stage: str) -> None:

        if stage == "fit":
            split_names = [self.train_split, self.val_split]
        elif stage == "validate":
            split_names = [self.val_split]
        elif stage == "test":
            split_names = [self.test_split]
        else:
            raise NotImplementedError(f"not implemented for stage={stage} ")

        for split in split_names:
            if split is None or split not in self.dataset:
                continue

            self._data[split] = self.encode_documents(self.dataset[split])

    def data_split(self, split: Optional[str] = None) -> Union[
        TaskEncodingDataset[TaskEncoding[DocumentType, InputEncoding, TargetEncoding]],
        IterableTaskEncodingDataset[TaskEncoding[DocumentType, InputEncoding, TargetEncoding]],
    ]:
        if split is None or split not in self._data:
            raise ValueError(f"data for split={split} not available")
        return self._data[split]

    def train_dataloader(
        self,
    ) -> DataLoader[TaskEncoding[DocumentType, InputEncoding, TargetEncoding]]:
        ds = self.data_split(self.train_split)
        return DataLoader(
            dataset=ds,
            collate_fn=self.taskmodule.collate,
            # don't shuffle streamed datasets
            shuffle=not isinstance(ds, IterableTaskEncodingDataset),
            **self.dataloader_kwargs,
        )

    def val_dataloader(
        self,
    ) -> DataLoader[TaskEncoding[DocumentType, InputEncoding, TargetEncoding]]:
        return DataLoader(
            dataset=self.data_split(self.val_split),
            collate_fn=self.taskmodule.collate,
            shuffle=False,
            **self.dataloader_kwargs,
        )

    def test_dataloader(
        self,
    ) -> DataLoader[TaskEncoding[DocumentType, InputEncoding, TargetEncoding]]:
        return DataLoader(
            dataset=self.data_split(self.test_split),
            collate_fn=self.taskmodule.collate,
            shuffle=False,
            **self.dataloader_kwargs,
        )



import logging
from collections import defaultdict
from functools import partial
from typing import Callable, Collection, Dict, Hashable, Optional, Tuple, Union

import pandas as pd
from pie_core import Annotation, Document, DocumentMetric

from pytorch_ie.utils.hydra import resolve_target

logger = logging.getLogger(__name__)


def has_one_of_the_labels(ann: Annotation, label_field: str, labels: Collection[str]) -> bool:
    return getattr(ann, label_field) in labels


def has_this_label(ann: Annotation, label_field: str, label: str) -> bool:
    return getattr(ann, label_field) == label


class F1Metric(DocumentMetric):
    """Computes the (micro aggregated) F1 score for a given layer. If labels are provided,
    it also computes the F1 score for each label separately and the macro F1 score.

    Args:
        layer: The layer to compute the F1 score for.
        labels: If provided, calculate F1 score for each label.
        label_field: The field to use for the label. Defaults to "label".
        show_as_markdown: If True, logs the F1 score as markdown on the console when calling compute().
    """

    def __init__(
        self,
        layer: str,
        labels: Optional[Union[Collection[str], str]] = None,
        label_field: str = "label",
        show_as_markdown: bool = False,
        annotation_processor: Optional[Union[Callable[[Annotation], Hashable], str]] = None,
    ):
        super().__init__()
        self.layer = layer
        self.label_field = label_field
        self.show_as_markdown = show_as_markdown
        self.annotation_processor: Optional[Callable[[Annotation], Hashable]]
        if isinstance(annotation_processor, str):
            self.annotation_processor = resolve_target(annotation_processor)
        else:
            self.annotation_processor = annotation_processor

        self.per_label = labels is not None
        self.infer_labels = False
        if self.per_label:
            if isinstance(labels, str):
                if labels != "INFERRED":
                    raise ValueError(
                        "labels can only be 'INFERRED' if per_label is True and labels is a string"
                    )
                self.labels = []
                self.infer_labels = True
            elif isinstance(labels, Collection):
                if not all(isinstance(label, str) for label in labels):
                    raise ValueError("labels must be a collection of strings")
                if "MICRO" in labels or "MACRO" in labels:
                    raise ValueError(
                        "labels cannot contain 'MICRO' or 'MACRO' because they are used to capture aggregated metrics"
                    )
                if len(labels) == 0:
                    raise ValueError("labels cannot be empty")
                self.labels = list(labels)
            else:
                raise ValueError("labels must be a string or a collection of strings")

    def reset(self):
        self.counts = defaultdict(lambda: (0, 0, 0))

    def calculate_counts(
        self,
        document: Document,
        annotation_filter: Optional[Callable[[Annotation], bool]] = None,
        annotation_processor: Optional[Callable[[Annotation], Hashable]] = None,
    ) -> Tuple[int, int, int]:
        annotation_processor = annotation_processor or (lambda ann: ann)
        annotation_filter = annotation_filter or (lambda ann: True)
        predicted_annotations = {
            annotation_processor(ann)
            for ann in document[self.layer].predictions
            if annotation_filter(ann)
        }
        gold_annotations = {
            annotation_processor(ann) for ann in document[self.layer] if annotation_filter(ann)
        }
        tp = len([ann for ann in predicted_annotations & gold_annotations])
        fn = len([ann for ann in gold_annotations - predicted_annotations])
        fp = len([ann for ann in predicted_annotations - gold_annotations])
        return tp, fp, fn

    def add_counts(self, counts: Tuple[int, int, int], label: str):
        self.counts[label] = (
            self.counts[label][0] + counts[0],
            self.counts[label][1] + counts[1],
            self.counts[label][2] + counts[2],
        )

    def _update(self, document: Document):
        new_counts = self.calculate_counts(
            document=document,
            annotation_filter=(
                partial(has_one_of_the_labels, label_field=self.label_field, labels=self.labels)
                if self.per_label and not self.infer_labels
                else None
            ),
            annotation_processor=self.annotation_processor,
        )
        self.add_counts(new_counts, label="MICRO")
        if self.infer_labels:
            layer = document[self.layer]
            # collect labels from gold data and predictions
            for ann in list(layer) + list(layer.predictions):
                label = getattr(ann, self.label_field)
                if label not in self.labels:
                    self.labels.append(label)
        if self.per_label:
            for label in self.labels:
                new_counts = self.calculate_counts(
                    document=document,
                    annotation_filter=partial(
                        has_this_label, label_field=self.label_field, label=label
                    ),
                    annotation_processor=self.annotation_processor,
                )
                self.add_counts(new_counts, label=label)

    def _compute(self) -> Dict[str, Dict[str, float]]:
        res = dict()
        if self.per_label:
            res["MACRO"] = {"f1": 0.0, "p": 0.0, "r": 0.0}
        for label, counts in self.counts.items():
            tp, fp, fn = counts
            if tp == 0:
                p, r, f1 = 0.0, 0.0, 0.0
            else:
                p = tp / (tp + fp)
                r = tp / (tp + fn)
                f1 = 2 * p * r / (p + r)
            res[label] = {"f1": f1, "p": p, "r": r, "s": tp + fn}
            if self.per_label and label in self.labels:
                res["MACRO"]["f1"] += f1 / len(self.labels)
                res["MACRO"]["p"] += p / len(self.labels)
                res["MACRO"]["r"] += r / len(self.labels)
        if self.show_as_markdown:
            logger.info(f"\n{self.layer}:\n{pd.DataFrame(res).round(3).T.to_markdown()}")
        return res



from .simple_transformer_text_classification import SimpleTransformerTextClassificationTaskModule
from .transformer_re_text_classification import TransformerRETextClassificationTaskModule
from .transformer_seq2seq import TransformerSeq2SeqTaskModule
from .transformer_span_classification import TransformerSpanClassificationTaskModule
from .transformer_text_classification import TransformerTextClassificationTaskModule
from .transformer_token_classification import TransformerTokenClassificationTaskModule

__all__ = [
    "SimpleTransformerTextClassificationTaskModule",
    "TransformerRETextClassificationTaskModule",
    "TransformerSeq2SeqTaskModule",
    "TransformerSpanClassificationTaskModule",
    "TransformerTextClassificationTaskModule",
    "TransformerTokenClassificationTaskModule",
]



from pytorch_ie.models.transformer_seq2seq import TransformerSeq2SeqModel
from pytorch_ie.models.transformer_span_classification import TransformerSpanClassificationModel
from pytorch_ie.models.transformer_text_classification import TransformerTextClassificationModel
from pytorch_ie.models.transformer_token_classification import TransformerTokenClassificationModel

__all__ = [
    "TransformerSeq2SeqModel",
    "TransformerSpanClassificationModel",
    "TransformerTextClassificationModel",
    "TransformerTokenClassificationModel",
]



from collections.abc import Iterator, Sequence
from typing import TypeVar, Union, overload

import torch.utils.data.dataset as torch_dataset
from pie_core import TaskEncoding

TaskEncodingType = TypeVar("TaskEncodingType", bound=TaskEncoding)


class TaskEncodingDataset(torch_dataset.Dataset[TaskEncodingType]):
    def __init__(self, encodings: Sequence[TaskEncodingType]):
        self._encodings = encodings

    @overload
    def __getitem__(self, index: int) -> TaskEncodingType: ...

    @overload
    def __getitem__(self, s: slice) -> Sequence[TaskEncodingType]: ...

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[TaskEncodingType, Sequence[TaskEncodingType]]:
        return self._encodings[index]

    def __len__(self):
        return len(self._encodings)


class IterableTaskEncodingDataset(torch_dataset.IterableDataset[TaskEncodingType]):
    def __iter__(self) -> Iterator[TaskEncodingType]:
        yield from self._encodings

    def __init__(self, encodings: Iterator[TaskEncodingType]):
        self._encodings = encodings



import torch.nn.functional as F
from torch import nn


class MLP(nn.Module):
    """Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(
            nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim])
        )

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x



# kept for backward compatibility
from pie_core import AutoTaskModule

# kept for backward compatibility
from pytorch_ie.model import AutoPyTorchIEModel as AutoModel
from pytorch_ie.pipeline import PyTorchIEPipeline as AutoPipeline



import logging
from collections import defaultdict
from typing import Callable, Dict, Optional, Tuple, Union

import pandas as pd
from pie_core import Annotation, Document, DocumentMetric

from pytorch_ie.utils.hydra import resolve_target

logger = logging.getLogger(__name__)


class ConfusionMatrix(DocumentMetric):
    """Computes the confusion matrix for a given annotation layer that contains labeled annotations.

    Args:
        layer: The layer to compute the confusion matrix for.
        label_field: The field to use for the label. Defaults to "label".
        unassignable_label: The label to use for false negative annotations. Defaults to "UNASSIGNABLE".
        undetected_label: The label to use for false positive annotations. Defaults to "UNDETECTED".
        strict: If True, raises an error if a base annotation has multiple gold labels. If False, logs a warning.
        show_as_markdown: If True, logs the confusion matrix as markdown on the console when calling compute().
        annotation_processor: A callable that processes the annotations before calculating the confusion matrix.
    """

    def __init__(
        self,
        layer: str,
        label_field: str = "label",
        show_as_markdown: bool = False,
        unassignable_label: str = "UNASSIGNABLE",
        undetected_label: str = "UNDETECTED",
        strict: bool = True,
        annotation_processor: Optional[Union[Callable[[Annotation], Annotation], str]] = None,
    ):
        super().__init__()
        self.layer = layer
        self.label_field = label_field
        self.unassignable_label = unassignable_label
        self.undetected_label = undetected_label
        self.strict = strict
        self.show_as_markdown = show_as_markdown
        self.annotation_processor: Optional[Callable[[Annotation], Annotation]]
        if isinstance(annotation_processor, str):
            self.annotation_processor = resolve_target(annotation_processor)
        else:
            self.annotation_processor = annotation_processor

    def reset(self):
        self.counts: Dict[Tuple[str, str], int] = defaultdict(int)

    def calculate_counts(
        self,
        document: Document,
        annotation_filter: Optional[Callable[[Annotation], bool]] = None,
        annotation_processor: Optional[Callable[[Annotation], Annotation]] = None,
    ) -> Dict[Tuple[str, str], int]:
        annotation_processor = annotation_processor or (lambda ann: ann)
        annotation_filter = annotation_filter or (lambda ann: True)
        predicted_annotations = {
            annotation_processor(ann)
            for ann in document[self.layer].predictions
            if annotation_filter(ann)
        }
        gold_annotations = {
            annotation_processor(ann) for ann in document[self.layer] if annotation_filter(ann)
        }
        base2gold = defaultdict(list)
        for ann in gold_annotations:
            base_ann_kwargs = {self.label_field: "DUMMY_LABEL"}
            base_ann = ann.copy(**base_ann_kwargs)
            base2gold[base_ann].append(ann)
        base2pred = defaultdict(list)
        for ann in predicted_annotations:
            base_ann_kwargs = {self.label_field: "DUMMY_LABEL"}
            base_ann = ann.copy(**base_ann_kwargs)
            base2pred[base_ann].append(ann)

        # (gold_label, pred_label) -> count
        counts: Dict[Tuple[str, str], int] = defaultdict(int)
        for base_ann in set(base2gold) | set(base2pred):
            gold_labels = [getattr(ann, self.label_field) for ann in base2gold[base_ann]]
            pred_labels = [getattr(ann, self.label_field) for ann in base2pred[base_ann]]

            if self.undetected_label in gold_labels:
                raise ValueError(
                    f"The gold annotation has the label '{self.undetected_label}' for undetected instances. "
                    f"Set a different undetected_label."
                )
            if self.unassignable_label in pred_labels:
                raise ValueError(
                    f"The predicted annotation has the label '{self.unassignable_label}' for unassignable predictions. "
                    f"Set a different unassignable_label."
                )

            if len(gold_labels) > 1:
                msg = f"The base annotation {base_ann} has multiple gold labels: {gold_labels}."
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.warning(msg + " Skip this base annotation.")
                    continue

            # use placeholder labels for empty gold or prediction labels
            if len(gold_labels) == 0:
                gold_labels.append(self.undetected_label)
            if len(pred_labels) == 0:
                pred_labels.append(self.unassignable_label)

            # main logic
            for gold_label in gold_labels:
                for pred_label in pred_labels:
                    counts[(gold_label, pred_label)] += 1

        return counts

    def add_counts(self, counts: Dict[Tuple[str, str], int]):
        for key, value in counts.items():
            self.counts[key] += value

    def _update(self, document: Document):
        new_counts = self.calculate_counts(
            document=document,
            annotation_processor=self.annotation_processor,
        )
        self.add_counts(new_counts)

    def _compute(self) -> Dict[str, Dict[str, int]]:

        res: Dict[str, Dict[str, int]] = defaultdict(dict)
        for gold_label, pred_label in sorted(self.counts):
            res[gold_label][pred_label] = self.counts[(gold_label, pred_label)]

        if self.show_as_markdown:
            res_df = pd.DataFrame(res).fillna(0)
            # index is prediction, columns is gold
            gold_labels = res_df.columns
            pred_labels = res_df.index

            # re-arrange index and columns: sort and put undetected_label and unassignable_label at the end
            gold_labels_sorted = sorted(
                [gold_label for gold_label in gold_labels if gold_label != self.undetected_label]
            )
            # re-add undetected_label at the end, if it was in the gold labels
            if self.undetected_label in gold_labels:
                gold_labels_sorted = gold_labels_sorted + [self.undetected_label]
            pred_labels_sorted = sorted(
                [pred_label for pred_label in pred_labels if pred_label != self.unassignable_label]
            )
            # re-add unassignable_label at the end, if it was in the pred labels
            if self.unassignable_label in pred_labels:
                pred_labels_sorted = pred_labels_sorted + [self.unassignable_label]
            res_df_sorted = res_df.loc[pred_labels_sorted, gold_labels_sorted]

            # transpose and show as markdown: index is now gold, columns is prediction
            logger.info(f"\n{self.layer}:\n{res_df_sorted.T.to_markdown()}")
        return res



from typing import Any, Tuple

import torch
from transformers import AutoConfig, AutoModelForSeq2SeqLM, BatchEncoding
from transformers.modeling_outputs import Seq2SeqLMOutput
from typing_extensions import TypeAlias

from pytorch_ie.model import PyTorchIEModel
from pytorch_ie.models.interface import RequiresModelNameOrPath

ModelInputType: TypeAlias = BatchEncoding
ModelOutputType: TypeAlias = Seq2SeqLMOutput

ModelStepInputType: TypeAlias = Tuple[ModelInputType]


@PyTorchIEModel.register()
class TransformerSeq2SeqModel(PyTorchIEModel, RequiresModelNameOrPath):
    def __init__(self, model_name_or_path: str, learning_rate: float = 1e-5, **kwargs) -> None:
        super().__init__(**kwargs)

        self.save_hyperparameters()

        self.learning_rate = learning_rate

        if self.is_from_pretrained:
            config = AutoConfig.from_pretrained(model_name_or_path)
            self.model = AutoModelForSeq2SeqLM.from_config(config=config)
        else:
            self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name_or_path)

    def forward(self, inputs: ModelInputType) -> ModelOutputType:
        return self.model(**inputs)

    def predict(
        self,
        inputs: Any,
        **kwargs,
    ) -> Any:
        if "labels" in inputs:
            inputs = {k: v for k, v in inputs.items() if k != "labels"}

        return self.model.generate(**inputs, **kwargs)

    def step(self, batch: ModelStepInputType):
        inputs = batch[0]
        output = self.forward(inputs)

        loss = output.loss

        return loss

    def training_step(self, batch: ModelStepInputType, batch_idx):
        loss = self.step(batch)
        self.log("train/loss", loss, on_step=True, on_epoch=False, prog_bar=True)

        return loss

    def validation_step(self, batch: ModelStepInputType, batch_idx):
        loss = self.step(batch)
        self.log("val/loss", loss, on_step=False, on_epoch=True, prog_bar=True)

        return loss

    def test_step(self, batch: ModelStepInputType, batch_idx):
        loss = self.step(batch)
        self.log("test/loss", loss, on_step=False, on_epoch=True, prog_bar=True)

        return loss

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.learning_rate)



"""
workflow:
    document
        -> (input_encoding, target_encoding) -> task_encoding
            -> model_encoding -> model_output
        -> task_output
    -> document
"""

import logging
from typing import Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Type, Union

import numpy as np
import torch
import torch.nn.functional as F
from pie_core import TaskEncoding, TaskModule
from transformers import AutoTokenizer
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import BatchEncoding, TruncationStrategy
from typing_extensions import TypeAlias

from pytorch_ie.annotations import LabeledSpan, MultiLabeledSpan, Span
from pytorch_ie.documents import (
    TextDocument,
    TextDocumentWithLabeledSpans,
    TextDocumentWithLabeledSpansAndLabeledPartitions,
    TextDocumentWithLabeledSpansAndSentences,
)
from pytorch_ie.models.transformer_span_classification import ModelOutputType, ModelStepInputType

InputEncodingType: TypeAlias = BatchEncoding
TargetEncodingType: TypeAlias = Sequence[Tuple[int, int, int]]

TaskEncodingType: TypeAlias = TaskEncoding[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
]
TaskOutputType: TypeAlias = Dict[str, Any]

TaskModuleType: TypeAlias = TaskModule[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
    ModelStepInputType,
    ModelOutputType,
    TaskOutputType,
]


logger = logging.getLogger(__name__)


@TaskModule.register()
class TransformerSpanClassificationTaskModule(TaskModuleType):
    PREPARED_ATTRIBUTES = ["label_to_id"]

    def __init__(
        self,
        tokenizer_name_or_path: str,
        entity_annotation: str = "labeled_spans",
        single_sentence: bool = False,
        sentence_annotation: str = "sentences",
        padding: Union[bool, str, PaddingStrategy] = True,
        truncation: Union[bool, str, TruncationStrategy] = True,
        max_length: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        label_pad_token_id: int = -100,
        label_to_id: Optional[Dict[str, int]] = None,
        multi_label: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        if multi_label:
            raise NotImplementedError(
                "Multi-label classification (multi_label=True) is not supported yet."
            )

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)
        self.entity_annotation = entity_annotation
        self.single_sentence = single_sentence
        self.sentence_annotation = sentence_annotation
        if label_to_id is not None:
            self.label_to_id = label_to_id
        self.padding = padding
        self.truncation = truncation
        self.max_length = max_length
        self.pad_to_multiple_of = pad_to_multiple_of
        self.label_pad_token_id = label_pad_token_id
        self.multi_label = multi_label

    @property
    def document_type(self) -> Optional[Type[TextDocument]]:
        dt: Type[TextDocument]
        if self.single_sentence:
            dt = TextDocumentWithLabeledSpansAndSentences
        else:
            dt = TextDocumentWithLabeledSpans

        if self.entity_annotation == "labeled_spans":
            return dt
        else:
            logger.warning(
                f"entity_annotation={self.entity_annotation} is "
                f"not the default value ('labeled_spans'), so the taskmodule {type(self).__name__} can not request "
                f"the usual document type ({dt.__name__}) for auto-conversion because this has the bespoken default "
                f"value as layer name instead of the provided one."
            )
            return None

    def _prepare(self, documents: Sequence[TextDocument]) -> None:
        labels: Set[str] = set()
        for document in documents:
            entities: Union[Sequence[LabeledSpan], Sequence[MultiLabeledSpan]] = document[
                self.entity_annotation
            ]

            for entity in entities:
                if self.multi_label and not isinstance(entity, MultiLabeledSpan):
                    raise ValueError("Spans must be MultiLabeledSpan if multi_label=True.")

                if not self.multi_label and not isinstance(entity, LabeledSpan):
                    raise ValueError("Spans must be LabeledSpan if multi_label=False.")

                if self.multi_label:
                    labels.update(entity.label)
                else:
                    labels.add(entity.label)

        self.label_to_id = {"O": 0}
        label_id = 1
        for label in sorted(labels):
            self.label_to_id[label] = label_id
            label_id += 1

    def _post_prepare(self) -> None:
        self.id_to_label = {v: k for k, v in self.label_to_id.items()}

    def encode_input(
        self,
        document: TextDocument,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType]]]:
        partitions: Sequence[Span]
        if self.single_sentence:
            partitions = document[self.sentence_annotation]
        else:
            partitions = [Span(start=0, end=len(document.text))]

        task_encodings: List[TaskEncoding] = []
        for partition_idx, partition in enumerate(partitions):
            inputs = self.tokenizer(
                document.text[partition.start : partition.end],
                padding=False,
                truncation=self.truncation,
                max_length=self.max_length,
                is_split_into_words=False,
                return_offsets_mapping=True,
                return_special_tokens_mask=True,
            )

            metadata = {
                "offset_mapping": inputs.pop("offset_mapping"),
                "special_tokens_mask": inputs.pop("special_tokens_mask"),
            }

            if self.single_sentence:
                metadata["partition_idx"] = partition_idx

            task_encodings.append(
                TaskEncoding(
                    document=document,
                    inputs=inputs,
                    metadata=metadata,
                )
            )

        return task_encodings

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> TargetEncodingType:
        # inputs is a transformers BatchEncoding object
        document = task_encoding.document
        inputs = task_encoding.inputs
        metadata = task_encoding.metadata

        targets: List[Tuple[int, int, int]] = []
        entities: Sequence[LabeledSpan] = document[self.entity_annotation]
        if self.single_sentence:
            partition_idx = metadata["partition_idx"]
            partitions: Sequence[Span] = document[self.sentence_annotation]
            assert (
                partitions is not None
            ), f"document has no span annotations with name '{self.sentence_annotation}'"
            partition = partitions[partition_idx]

            for entity in entities:
                if entity.start < partition.start or entity.end > partition.end:
                    continue

                entity_start = entity.start - partition.start
                entity_end = entity.end - partition.start

                start_idx = inputs.char_to_token(entity_start)
                end_idx = inputs.char_to_token(entity_end - 1)

                # TODO: remove this is if case
                if start_idx is None or end_idx is None:
                    logger.warning(
                        f"Entity annotation does not start or end with a token, it will be skipped: {entity}"
                    )
                    continue

                targets.append((start_idx, end_idx, self.label_to_id[entity.label]))
        else:
            for entity in entities:
                start_idx = inputs.char_to_token(entity.start)
                end_idx = inputs.char_to_token(entity.end - 1)
                targets.append((start_idx, end_idx, self.label_to_id[entity.label]))

        return targets

    def unbatch_output(self, model_output: ModelOutputType) -> Sequence[TaskOutputType]:
        logits = model_output["logits"]
        probs = F.softmax(logits, dim=-1).detach().cpu().float().numpy()
        label_ids = torch.argmax(logits, dim=-1).detach().cpu().numpy()

        start_indices = model_output["start_indices"].detach().cpu().numpy()
        end_indices = model_output["end_indices"].detach().cpu().numpy()
        batch_indices = model_output["batch_indices"].detach().cpu().numpy()

        tags: List[List[Tuple[str, Tuple[int, int]]]] = [[] for _ in np.unique(batch_indices)]
        probabilities: List[List[float]] = [[] for _ in np.unique(batch_indices)]
        for start, end, batch_idx, label_id, prob in zip(
            start_indices, end_indices, batch_indices, label_ids, probs
        ):
            label = self.id_to_label[label_id]
            if label != "O":
                tags[batch_idx].append((label, (start, end)))
                probabilities[batch_idx].append(prob[label_id])

        return [{"tags": t, "probabilities": p} for t, p in zip(tags, probabilities)]

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, Union[LabeledSpan, MultiLabeledSpan]]]:
        document = task_encoding.document
        metadata = task_encoding.metadata

        if self.single_sentence:
            partitions: Sequence[Span] = document[self.sentence_annotation]
            sentence = partitions[metadata["sentence_index"]]

            spans = task_output["tags"]
            probabilities = task_output["probabilities"]
            for (label, (start, end)), probability in zip(spans, probabilities):
                yield (
                    self.entity_annotation,
                    LabeledSpan(
                        start=sentence.start + metadata["offset_mapping"][start][0],
                        end=sentence.start + metadata["offset_mapping"][end][1],
                        label=label,
                        score=float(probability),
                    ),
                )
        else:
            spans = task_output["tags"]
            probabilities = task_output["probabilities"]
            for (label, (start, end)), probability in zip(spans, probabilities):
                yield (
                    self.entity_annotation,
                    LabeledSpan(
                        start=metadata["offset_mapping"][start][0],
                        end=metadata["offset_mapping"][end][1],
                        label=label,
                        score=float(probability),
                    ),
                )

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> ModelStepInputType:
        input_features = [task_encoding.inputs for task_encoding in task_encodings]

        inputs: Dict[str, torch.Tensor] = self.tokenizer.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        if not task_encodings[0].has_targets:
            return inputs, None

        targets: Sequence[TargetEncodingType] = [
            task_encoding.targets for task_encoding in task_encodings
        ]

        inputs = {k: torch.tensor(v, dtype=torch.int64) for k, v in inputs.items()}

        return inputs, targets



import dataclasses
from typing import Any, Dict, Optional, Tuple

from pie_core import AnnotationLayer, Document, annotation_field
from typing_extensions import TypeAlias

from pytorch_ie.annotations import (
    BinaryRelation,
    Label,
    LabeledMultiSpan,
    LabeledSpan,
    MultiLabel,
    Span,
)


@dataclasses.dataclass
class WithMetadata:
    id: Optional[str] = None
    metadata: Dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class WithTokens:
    tokens: Tuple[str, ...]


@dataclasses.dataclass
class WithText:
    text: str


@dataclasses.dataclass
class TextBasedDocument(WithMetadata, WithText, Document):
    pass


@dataclasses.dataclass
class TokenBasedDocument(WithMetadata, WithTokens, Document):
    def __post_init__(self) -> None:

        # When used in a dataset, the document gets serialized to json like structure which does not know tuples,
        # so they get converted to lists. This is a workaround to automatically convert the "tokens" back to tuples
        # when the document is created from a dataset.
        if isinstance(self.tokens, list):
            object.__setattr__(self, "tokens", tuple(self.tokens))
        elif not isinstance(self.tokens, tuple):
            raise ValueError("tokens must be a tuple.")

        # Call the default document construction code
        super().__post_init__()


# backwards compatibility
TextDocument: TypeAlias = TextBasedDocument


@dataclasses.dataclass
class DocumentWithLabel(Document):
    label: AnnotationLayer[Label] = annotation_field()


@dataclasses.dataclass
class DocumentWithMultiLabel(Document):
    label: AnnotationLayer[MultiLabel] = annotation_field()


@dataclasses.dataclass
class TextDocumentWithLabel(DocumentWithLabel, TextBasedDocument):
    pass


@dataclasses.dataclass
class TextDocumentWithMultiLabel(DocumentWithMultiLabel, TextBasedDocument):
    pass


@dataclasses.dataclass
class TextDocumentWithLabeledPartitions(TextBasedDocument):
    labeled_partitions: AnnotationLayer[LabeledSpan] = annotation_field(target="text")


@dataclasses.dataclass
class TextDocumentWithSentences(TextBasedDocument):
    sentences: AnnotationLayer[Span] = annotation_field(target="text")


@dataclasses.dataclass
class TextDocumentWithSpans(TextBasedDocument):
    spans: AnnotationLayer[Span] = annotation_field(target="text")


@dataclasses.dataclass
class TextDocumentWithLabeledSpans(TextBasedDocument):
    labeled_spans: AnnotationLayer[LabeledSpan] = annotation_field(target="text")


@dataclasses.dataclass
class TextDocumentWithLabeledSpansAndLabeledPartitions(
    TextDocumentWithLabeledSpans, TextDocumentWithLabeledPartitions
):
    pass


@dataclasses.dataclass
class TextDocumentWithLabeledSpansAndSentences(
    TextDocumentWithLabeledSpans, TextDocumentWithSentences
):
    pass


@dataclasses.dataclass
class TextDocumentWithLabeledSpansAndBinaryRelations(TextDocumentWithLabeledSpans):
    binary_relations: AnnotationLayer[BinaryRelation] = annotation_field(target="labeled_spans")


@dataclasses.dataclass
class TextDocumentWithLabeledSpansBinaryRelationsAndLabeledPartitions(
    TextDocumentWithLabeledSpansAndLabeledPartitions,
    TextDocumentWithLabeledSpansAndBinaryRelations,
    TextDocumentWithLabeledPartitions,
):
    pass


@dataclasses.dataclass
class TextDocumentWithSpansAndBinaryRelations(TextDocumentWithSpans):
    binary_relations: AnnotationLayer[BinaryRelation] = annotation_field(target="spans")


@dataclasses.dataclass
class TextDocumentWithSpansAndLabeledPartitions(
    TextDocumentWithSpans, TextDocumentWithLabeledPartitions
):
    pass


@dataclasses.dataclass
class TextDocumentWithSpansBinaryRelationsAndLabeledPartitions(
    TextDocumentWithSpansAndLabeledPartitions,
    TextDocumentWithSpansAndBinaryRelations,
    TextDocumentWithLabeledPartitions,
):
    pass


@dataclasses.dataclass
class TextDocumentWithLabeledMultiSpans(TextBasedDocument):
    labeled_multi_spans: AnnotationLayer[LabeledMultiSpan] = annotation_field(target="text")


@dataclasses.dataclass
class TextDocumentWithLabeledMultiSpansAndLabeledPartitions(
    TextDocumentWithLabeledMultiSpans, TextDocumentWithLabeledPartitions
):
    pass


@dataclasses.dataclass
class TextDocumentWithLabeledMultiSpansAndBinaryRelations(TextDocumentWithLabeledMultiSpans):
    binary_relations: AnnotationLayer[BinaryRelation] = annotation_field(
        target="labeled_multi_spans"
    )


@dataclasses.dataclass
class TextDocumentWithLabeledMultiSpansBinaryRelationsAndLabeledPartitions(
    TextDocumentWithLabeledMultiSpansAndLabeledPartitions,
    TextDocumentWithLabeledMultiSpansAndBinaryRelations,
):
    pass



"""
workflow:
    Document
        -> (InputEncoding, TargetEncoding) -> TaskEncoding -> TaskBatchEncoding
            -> ModelBatchEncoding -> ModelBatchOutput
        -> TaskOutput
    -> Document
"""

import copy
import logging
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple, Type, Union

import torch
import torch.nn.functional as F
from pie_core import TaskEncoding, TaskModule
from transformers import AutoTokenizer
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import BatchEncoding, TruncationStrategy
from typing_extensions import TypeAlias

from pytorch_ie.annotations import LabeledSpan, Span
from pytorch_ie.documents import (
    TextDocument,
    TextDocumentWithLabeledSpans,
    TextDocumentWithLabeledSpansAndLabeledPartitions,
)
from pytorch_ie.models.transformer_token_classification import ModelOutputType, ModelStepInputType
from pytorch_ie.utils.span import (
    bio_tags_to_spans,
    convert_span_annotations_to_tag_sequence,
    get_char_to_token_mapper,
    get_special_token_mask,
    has_overlap,
)
from pytorch_ie.utils.window import enumerate_windows

InputEncodingType: TypeAlias = Union[Dict[str, Any], BatchEncoding]
TargetEncodingType: TypeAlias = Sequence[int]

TaskEncodingType: TypeAlias = TaskEncoding[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
]
TaskOutputType: TypeAlias = Dict[str, Any]

TaskModuleType: TypeAlias = TaskModule[
    TextDocument,
    InputEncodingType,
    TargetEncodingType,
    ModelStepInputType,
    ModelOutputType,
    TaskOutputType,
]

logger = logging.getLogger(__name__)


@TaskModule.register()
class TransformerTokenClassificationTaskModule(TaskModuleType):
    PREPARED_ATTRIBUTES = ["label_to_id"]

    def __init__(
        self,
        tokenizer_name_or_path: str,
        entity_annotation: str = "labeled_spans",
        partition_annotation: Optional[str] = None,
        padding: Union[bool, str, PaddingStrategy] = True,
        truncation: Union[bool, str, TruncationStrategy] = False,
        max_length: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        label_pad_token_id: int = -100,
        label_to_id: Optional[Dict[str, int]] = None,
        max_window: Optional[int] = None,
        window_overlap: int = 0,
        show_statistics: bool = False,
        include_ill_formed_predictions: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)
        self.entity_annotation = entity_annotation
        self.partition_annotation = partition_annotation
        if label_to_id is not None:
            self.label_to_id = label_to_id
        self.padding = padding
        self.truncation = truncation
        self.max_length = max_length
        self.pad_to_multiple_of = pad_to_multiple_of
        self.label_pad_token_id = label_pad_token_id
        self.max_window = max_window
        self.window_overlap = window_overlap
        self.show_statistics = show_statistics
        self.include_ill_formed_predictions = include_ill_formed_predictions

    @property
    def document_type(self) -> Optional[Type[TextDocument]]:
        dt: Type[TextDocument]
        if self.partition_annotation is not None:
            dt = TextDocumentWithLabeledSpansAndLabeledPartitions
        else:
            dt = TextDocumentWithLabeledSpans
        if self.entity_annotation == "labeled_spans":
            return dt
        else:
            logger.warning(
                f"entity_annotation={self.entity_annotation} is "
                f"not the default value ('labeled_spans'), so the taskmodule {type(self).__name__} can not request "
                f"the usual document type ({dt.__name__}) for auto-conversion because this has the bespoken default "
                f"value as layer name instead of the provided one."
            )
            return None

    def _prepare(self, documents: Sequence[TextDocument]) -> None:
        labels = set()
        for document in documents:
            entities: Sequence[LabeledSpan] = document[self.entity_annotation]

            for entity in entities:
                labels.add(entity.label)

        self.label_to_id = {"O": 0}
        current_id = 1
        for label in sorted(labels):
            for prefix in ["B", "I"]:
                self.label_to_id[f"{prefix}-{label}"] = current_id
                current_id += 1

    def _post_prepare(self):
        self.id_to_label = {v: k for k, v in self.label_to_id.items()}

    def encode_text(
        self, text, partition: Optional[Span] = None, add_special_tokens: bool = True
    ) -> BatchEncoding:
        if self.partition_annotation is not None and partition is None:
            raise ValueError(f"partitioning is enabled, but no partition is provided")

        text_partition = text[partition.start : partition.end] if partition is not None else text
        return self.tokenizer(
            text_partition,
            padding=False,
            truncation=False,
            max_length=None,
            is_split_into_words=False,
            return_offsets_mapping=True,
            return_special_tokens_mask=True,
            add_special_tokens=add_special_tokens,
        )

    def encode_input(
        self,
        document: TextDocument,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType]]]:
        partitions: Sequence[Optional[Span]]
        if self.partition_annotation is not None:
            partitions = document[self.partition_annotation]
        else:
            partitions = [None]

        task_encodings: List[TaskEncodingType] = []
        for partition_index, partition in enumerate(partitions):
            add_special_tokens = self.max_window is None
            inputs: BatchEncoding = self.encode_text(
                text=document.text, partition=partition, add_special_tokens=add_special_tokens
            )

            metadata = {
                "offset_mapping": inputs.pop("offset_mapping"),
                "special_tokens_mask": inputs.pop("special_tokens_mask"),
                "char_to_token_mapper": inputs.char_to_token,
            }

            if partition is not None:
                metadata["sentence_index"] = partition_index

            if self.max_window is None:
                task_encodings.append(
                    TaskEncoding(
                        document=document,
                        inputs=inputs,
                        metadata=metadata,
                    )
                )
            else:
                offset_mapping = metadata.pop("offset_mapping")
                # The actual number of tokens will be lower than max_window because we add the default special
                # tokens later on (e.g. CLS and SEP).
                max_window = self.max_window - self.tokenizer.num_special_tokens_to_add()
                token_ids = inputs["input_ids"]
                for token_slice, label_offset_slice in enumerate_windows(
                    sequence=token_ids, max_size=max_window, overlap=self.window_overlap
                ):
                    start_idx, end_idx = token_slice
                    new_input_ids = self.tokenizer.build_inputs_with_special_tokens(
                        token_ids_0=token_ids[start_idx:end_idx]
                    )
                    new_special_tokens_mask = get_special_token_mask(
                        token_ids_0=new_input_ids, tokenizer=self.tokenizer
                    )
                    window_inputs = {"input_ids": new_input_ids}
                    # for now, we copy just to keep "sentence_index"
                    window_metadata = copy.deepcopy(metadata)
                    window_metadata["special_tokens_mask"] = new_special_tokens_mask
                    offset_mapping_without_special_tokens = offset_mapping[start_idx:end_idx]
                    j = 0
                    current_offset_mapping: List[Tuple[int, int]] = []
                    # this maps from positions without special tokens to positions with special tokens
                    position_with_special_tokens = {}
                    for i, is_special_token in enumerate(new_special_tokens_mask):
                        if is_special_token:
                            current_offset_mapping.append((0, 0))
                        else:
                            position_with_special_tokens[j] = i
                            current_offset_mapping.append(offset_mapping_without_special_tokens[j])
                            j += 1
                    window_metadata["offset_mapping"] = current_offset_mapping
                    char_to_token_mapping: Dict[int, int] = {}
                    for token_idx, (char_start, char_end) in enumerate(current_offset_mapping):
                        for char_idx in range(char_start, char_end):
                            char_to_token_mapping[char_idx] = token_idx
                    window_metadata["char_to_token_mapper"] = get_char_to_token_mapper(
                        char_to_token_mapping=char_to_token_mapping,
                        char_start=offset_mapping_without_special_tokens[0][0],
                        char_end=offset_mapping_without_special_tokens[-1][1],
                    )
                    # new_metadata["window_tokens"] = token_slice
                    window_metadata["window_labels"] = (
                        position_with_special_tokens[label_offset_slice[0]],
                        # we have to look up the actual index, not the pythonic end position
                        position_with_special_tokens[label_offset_slice[1] - 1] + 1,
                    )

                    task_encodings.append(
                        TaskEncoding(
                            document=document,
                            inputs=window_inputs,
                            metadata=window_metadata,
                        )
                    )

        return task_encodings

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> Optional[TargetEncodingType]:
        metadata = task_encoding.metadata
        document = task_encoding.document

        entities: Sequence[LabeledSpan] = document[self.entity_annotation]

        partition = None
        if self.partition_annotation is not None:
            partition_index = metadata["sentence_index"]
            partitions = document[self.partition_annotation]
            partition = partitions[partition_index]
        tag_sequence_or_none = convert_span_annotations_to_tag_sequence(
            spans=entities,
            special_tokens_mask=metadata["special_tokens_mask"],
            char_to_token_mapper=metadata["char_to_token_mapper"],
            partition=partition,
            statistics=None,
        )
        if tag_sequence_or_none is None:
            logger.warning(
                f"can not create targets for document with id: {getattr(document, 'id', None)}, skip it"
            )
            return None
        else:
            tag_sequence = tag_sequence_or_none

        # exclude labels that are out of the window (when overlap is used)
        window_labels = metadata.get("window_labels")
        if window_labels is not None:
            tag_sequence[0 : window_labels[0]] = [None] * window_labels[0]
            tag_sequence[window_labels[1] :] = [None] * len(tag_sequence[window_labels[1] :])

        targets = [
            self.label_to_id[tag] if tag is not None else self.label_pad_token_id
            for tag in tag_sequence
        ]

        return targets

    def unbatch_output(self, model_output: ModelOutputType) -> Sequence[TaskOutputType]:
        logits = model_output["logits"]
        probabilities = F.softmax(logits, dim=-1).detach().cpu().float().numpy()
        indices = torch.argmax(logits, dim=-1).detach().cpu().numpy()
        tags = [[self.id_to_label[e] for e in b] for b in indices]
        return [{"tags": t, "probabilities": p} for t, p in zip(tags, probabilities)]

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, LabeledSpan]]:
        offset = 0
        if self.partition_annotation is not None:
            partitions = task_encoding.document[self.partition_annotation]
            offset = partitions[task_encoding.metadata["sentence_index"]].start

        tag_sequence = [
            "O" if is_special_token else tag
            for tag, is_special_token in zip(
                task_output["tags"], task_encoding.metadata["special_tokens_mask"]
            )
        ]

        spans = bio_tags_to_spans(
            tag_sequence, include_ill_formed=self.include_ill_formed_predictions
        )
        for label, (start, end) in spans:
            if "window_labels" in task_encoding.metadata:
                # Take only spans into account that are at least partly in the window. The model was not
                # trained to correctly predict spans that are just in the context.
                # NOTE: The "end" index is exclusive, but encoding.metadata["window_labels"][1] is inclusive!
                if not has_overlap((start, end + 1), task_encoding.metadata["window_labels"]):
                    continue
            yield (
                self.entity_annotation,
                LabeledSpan(
                    task_encoding.metadata["offset_mapping"][start][0] + offset,
                    task_encoding.metadata["offset_mapping"][end][1] + offset,
                    label,
                ),
            )

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> ModelStepInputType:
        input_features = [task_encoding.inputs for task_encoding in task_encodings]

        inputs = self.tokenizer.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        if not task_encodings[0].has_targets:
            return inputs, None

        target_list: List[TargetEncodingType] = [
            task_encoding.targets for task_encoding in task_encodings
        ]

        sequence_length = inputs["input_ids"].shape[1]
        padding_side = self.tokenizer.padding_side
        if padding_side == "right":
            target_list_padded = [
                list(t) + [self.label_pad_token_id] * (sequence_length - len(t))
                for t in target_list
            ]
        else:
            target_list_padded = [
                [self.label_pad_token_id] * (sequence_length - len(t)) + list(t)
                for t in target_list
            ]

        targets = torch.tensor(target_list_padded, dtype=torch.int64)

        return inputs, targets



# flake8: noqa

from pytorch_ie.auto import AutoModel, AutoPipeline, AutoTaskModule
from pytorch_ie.core import *
from pytorch_ie.datamodule import PieDataModule
from pytorch_ie.dataset import IterableTaskEncodingDataset, TaskEncodingDataset
from pytorch_ie.pipeline import PyTorchIEPipeline

# kept for backward compatibility
Pipeline = PyTorchIEPipeline



import sys

import pie_core
from pie_core import taskmodule
from pie_core.document import Annotation, AnnotationLayer, Document, annotation_field
from pie_core.metric import DocumentMetric
from pie_core.module_mixins import (
    EnterDatasetDictMixin,
    EnterDatasetMixin,
    ExitDatasetDictMixin,
    ExitDatasetMixin,
    WithDocumentTypeMixin,
)
from pie_core.preparable import PreparableMixin
from pie_core.statistic import DocumentStatistic
from pie_core.taskencoding import TaskEncoding, TaskEncodingSequence
from pie_core.taskmodule import TaskModule

from pytorch_ie import model
from pytorch_ie.dataset import IterableTaskEncodingDataset, TaskEncodingDataset
from pytorch_ie.model import PyTorchIEModel

submodules = ["document", "taskmodule", "metric", "statistic"]
for sub in submodules:
    module = getattr(pie_core, sub)
    sys.modules[f"{__name__}.{sub}"] = module

sys.modules[f"{__name__}.model"] = model

taskmodule.TaskEncodingDataset = TaskEncodingDataset
taskmodule.IterableTaskEncodingDataset = IterableTaskEncodingDataset

# backwards compatibility
AnnotationList = AnnotationLayer
RequiresDocumentTypeMixin = WithDocumentTypeMixin



from collections import defaultdict
from typing import Dict, Hashable, List, Optional, TypeVar

from pie_core import Document
from pie_core.document import BaseAnnotationList

from pytorch_ie.documents import WithMetadata


def deduplicate_annotation_dicts(
    annotation_dicts: List[Dict[str, Hashable]]
) -> List[Dict[str, Hashable]]:
    """Remove duplicate annotation dictionaries from a list of annotation dictionaries.

    Args:
        annotation_dicts: The list of annotation dictionaries to remove duplicates from.

    Returns:
        The list of annotation dictionaries with duplicates removed.
    """
    unique_annotation_dicts = []
    seen_annotation_dicts = set()
    for annotation_dict in annotation_dicts:
        annotation_dict_tuple = tuple(sorted(annotation_dict.items()))
        if annotation_dict_tuple not in seen_annotation_dicts:
            unique_annotation_dicts.append(annotation_dict)
            seen_annotation_dicts.add(annotation_dict_tuple)
    return unique_annotation_dicts


D = TypeVar("D", bound=Document)


def save_annotation_sources_to_metadata(
    document: D,
    annotation_id2source: Dict[int, List[str]],
    metadata_key: str,
    use_predictions: bool,
) -> None:
    """Save the source names for the annotations or predictions in the metadata of the document.

    Args:
        document: The document to save the source names in the metadata for.
        metadata_key: The key in the metadata where the source names should be stored.
        annotation_id2source: A mapping from annotation IDs to the source names. Should contain
            the ids of all annotations or predictions (depending on use_predictions) in the document.
        use_predictions: Whether to store the source names for the predictions or the annotations.
    """

    if not hasattr(document, "metadata"):
        raise ValueError("Document does not have metadata, can not store source names.")
    if metadata_key in document.metadata:
        raise ValueError(f"Metadata key '{metadata_key}' already exists in the document.")
    document.metadata[metadata_key] = defaultdict(dict)
    for annotation_field in document.annotation_fields():
        layer_name = annotation_field.name
        document.metadata[metadata_key][layer_name] = []
        layer: BaseAnnotationList
        if use_predictions:
            layer = document[layer_name].predictions
        else:
            layer = document[layer_name]
        for ann in layer:
            document.metadata[metadata_key][layer_name].append(annotation_id2source[ann._id])
    document.metadata[metadata_key] = dict(document.metadata[metadata_key])


def merge_annotations_from_documents(
    documents: Dict[str, D],
    metadata_key_source_annotations: Optional[str] = None,
    metadata_key_source_predictions: Optional[str] = None,
) -> D:
    """Merge annotations from multiple documents into a single document. Optionally, store the source
    names for all annotations / predictions in the metadata at key metadata_key_source_annotations
    / metadata_key_source_predictions, respectively.

    Note that this will remove any annotation duplicates.

    Args:
        documents: A dictionary mapping document source (e.g. dataset names) to documents.
        metadata_key_source_annotations: If not None, the key in the metadata where the source names
            for the (gold) annotations are stored.
        metadata_key_source_predictions: If not None, the key in the metadata where the source names
            for the predictions are stored.

    Returns:
        The document with merged annotations.
    """
    if len(documents) == 0:
        raise ValueError("No documents provided.")
    source_names = sorted(documents)
    first_source_name = source_names[0]
    merged_document: D = documents[first_source_name].copy(with_annotations=False)

    added_annotation_id2source_names: Dict[int, List[str]] = defaultdict(list)
    for source_name in source_names:
        document = documents[source_name]
        if type(document) is not type(merged_document):
            raise ValueError(
                f"Document types do not match: {type(document)} and {type(merged_document)}"
            )
        if isinstance(document, WithMetadata) and document.id is not None:
            if document.id != merged_document.id:
                raise ValueError(
                    f"Document IDs do not match: {document.id} and {merged_document.id}"
                )

        # Note: this does not check for duplicates!
        added_annotations = merged_document.add_all_annotations_from_other(
            other=document, strict=True
        )

        for layer_name, orig_id2new_annotation in added_annotations.items():
            for orig_id, new_annotation in orig_id2new_annotation.items():
                added_annotation_id2source_names[new_annotation._id].append(source_name)

    # this will remove duplicates. If duplicates have different scores, the one with the highest score will be kept
    merged_document = merged_document.deduplicate_annotations()

    # save source names in metadata (at key metadata_key_source_annotations / metadata_key_source_predictions
    #   for each layer in the order of the annotations / predictions)
    if metadata_key_source_annotations is not None:
        save_annotation_sources_to_metadata(
            document=merged_document,
            annotation_id2source=added_annotation_id2source_names,
            metadata_key=metadata_key_source_annotations,
            use_predictions=False,
        )
    if metadata_key_source_predictions is not None:
        save_annotation_sources_to_metadata(
            document=merged_document,
            annotation_id2source=added_annotation_id2source_names,
            metadata_key=metadata_key_source_predictions,
            use_predictions=True,
        )
    return merged_document



from typing import Any, Dict

import torch
from pie_core import Auto, Model
from pytorch_lightning import LightningModule


class PyTorchIEModel(Model, LightningModule):
    weights_file_name = "pytorch_model.bin"

    def _config(self) -> Dict[str, Any]:
        config = super()._config() or {}
        # add all hparams
        config.update(self.hparams)
        return config

    def save_model_file(self, model_file: str) -> None:
        """Save weights from a Pytorch model to a local directory."""
        model_to_save: torch.nn.Module = self.module if hasattr(self, "module") else self
        torch.save(model_to_save.state_dict(), model_file)

    def load_model_file(
        self, model_file: str, map_location: str = "cpu", strict: bool = False
    ) -> None:
        state_dict = torch.load(model_file, map_location=torch.device(map_location))
        self.load_state_dict(state_dict, strict=strict)
        # The model is set in evaluation mode by default using `model.eval()`
        # (dropout modules are deactivated). To train the model, you should first
        # set it back in training mode with `model.train()`. This is especially
        # important when using pytorch-lightning >= 2.2.0, as it maintains the
        # training/evaluation state of the model when training via `fit()`.
        self.eval()

    def decode(self, inputs: Any, outputs: Any) -> Any:
        return outputs

    def predict(self, inputs: Any, **kwargs) -> Any:
        outputs = self(inputs, **kwargs)
        decoded_outputs = self.decode(inputs=inputs, outputs=outputs)
        return decoded_outputs


# TODO: remove this class when all models are registered with @Model.register()
#   also see notes in PyTorchIEPipeline
class AutoPyTorchIEModel(Model, Auto[PyTorchIEModel]):

    BASE_CLASS = PyTorchIEModel



import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Type, Union

from pie_core import Document, DocumentStatistic
from transformers import AutoTokenizer, PreTrainedTokenizer

from pytorch_ie.documents import TextBasedDocument

logger = logging.getLogger(__name__)


class TokenCountCollector(DocumentStatistic):
    """Collects the token count of a field when tokenizing its content with a Huggingface
    tokenizer.

    The content of the field should be a string.
    """

    def __init__(
        self,
        tokenizer: Union[str, PreTrainedTokenizer],
        text_field: str = "text",
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        document_type: Optional[Type[Document]] = None,
        **kwargs,
    ):
        if document_type is None and text_field == "text":
            document_type = TextBasedDocument
        super().__init__(document_type=document_type, **kwargs)
        self.tokenizer = (
            AutoTokenizer.from_pretrained(tokenizer) if isinstance(tokenizer, str) else tokenizer
        )
        self.tokenizer_kwargs = tokenizer_kwargs or {}
        self.text_field = text_field

    def _collect(self, doc: Document) -> int:
        text = getattr(doc, self.text_field)
        encodings = self.tokenizer(text, **self.tokenizer_kwargs)
        tokens = encodings.tokens()
        return len(tokens)


class FieldLengthCollector(DocumentStatistic):
    """Collects the length of a field, e.g. to collect the number the characters in the input text.

    The field should be a list of sized elements.
    """

    def __init__(self, field: str, **kwargs):
        super().__init__(**kwargs)
        self.field = field

    def _collect(self, doc: Document) -> int:
        field_obj = getattr(doc, self.field)
        return len(field_obj)


class SubFieldLengthCollector(DocumentStatistic):
    """Collects the length of a subfield in a field, e.g. to collect the number of arguments of
    N-ary relations."""

    def __init__(self, field: str, subfield: str, **kwargs):
        super().__init__(**kwargs)
        self.field = field
        self.subfield = subfield

    def _collect(self, doc: Document) -> List[int]:
        field_obj = getattr(doc, self.field)
        lengths = []
        for entry in field_obj:
            subfield_obj = getattr(entry, self.subfield)
            lengths.append(len(subfield_obj))
        return lengths


class DummyCollector(DocumentStatistic):
    """A dummy collector that always returns 1, e.g. to count the number of documents.

    Can be used to count the number of documents.
    """

    DEFAULT_AGGREGATION_FUNCTIONS = ["sum"]

    def _collect(self, doc: Document) -> int:
        return 1


class LabelCountCollector(DocumentStatistic):
    """Collects the number of field entries per label, e.g. to collect the number of entities per
    type.

    The field should be a list of elements with a label attribute.

    Important: To make correct use of the result data, missing values need to be filled with 0, e.g.:
        {("ORG",): [2, 3], ("LOC",): [2]} -> {("ORG",): [2, 3], ("LOC",): [2, 0]}
    """

    DEFAULT_AGGREGATION_FUNCTIONS = ["mean", "std", "min", "max", "len", "sum"]

    def __init__(
        self, field: str, labels: Union[List[str], str], label_attribute: str = "label", **kwargs
    ):
        super().__init__(**kwargs)
        self.field = field
        self.label_attribute = label_attribute
        if not (isinstance(labels, list) or labels == "INFERRED"):
            raise ValueError("labels must be a list of strings or 'INFERRED'")
        if labels == "INFERRED":
            logger.warning(
                f"Inferring labels with {self.__class__.__name__} from data produces wrong results "
                f"for certain aggregation functions (e.g. 'mean', 'std', 'min') because zero values "
                f"are not included in the calculation. We remove these aggregation functions from "
                f"this collector, but be aware that the results may be wrong for your own aggregation "
                f"functions that rely on zero values."
            )
            self.aggregation_functions: Dict[str, Callable[[List], Any]] = {
                name: func
                for name, func in self.aggregation_functions.items()
                if name not in ["mean", "std", "min"]
            }

        self.labels = labels

    def _collect(self, doc: Document) -> Dict[str, int]:
        field_obj = getattr(doc, self.field)
        counts: Dict[str, int]
        if self.labels == "INFERRED":
            counts = defaultdict(int)
        else:
            counts = {label: 0 for label in self.labels}
        for elem in field_obj:
            label = getattr(elem, self.label_attribute)
            counts[label] += 1
        return dict(counts)
